from uio import *
from math import *
from urandom import *
from hpprime import *
import builtins

def gridstr(g):
  s=''
  for j in range(4):
    for i in range(4):
      if g[j][i]==0:
        s+='[o]'
      else:
        s+='[{0}]'.format(2**g[j][i].val)
    s+='\n'
  return s

colors=[
  0xEEE4DA,
  0xEDE0C8,
  0xF2B179,
  0xF59563,
  0xF67C5F,
  0xF65E3B,
  0xEDCF72,
  0xEDCC61,
  0xEDC850,
  0xEDC53F,
  0xEDC22E,
  0x3C3A3200
]

class Overlay():
  def __init__(self,opacity):
    self.opacity=int(opacity)

  def draw(self):
    pass

class GameOver(Overlay):
  def draw(self):
    #print(0.8*self.opacity)
    eval('rect_p(G1,rgb(255,255,255,{0}))'.format(255-0.75*self.opacity))
    w=eval('textout_p("Score: {0}",G4,0,0,4)'.format(board.score))
    eval('dimgrob_p(G3,320,240,#FFFFFF)')
    eval('textout_p("Game Over",G1,105,97,7,#776E65)')
    eval('textout_p("Score: {0}",G1,{1},120,4,#776E65)'.format(board.score,160-w//2))
    eval('textout_p("Touch to continue...",G1,103,220,3,#776E65)')
    #line(1,160,0,160,240,0)
    #eval('blit_p(G1,0,0,320,240,G3,0,0,320,240,#FFFFFF,{0})'.format(255-int(0.8*self.opacity)))

class Tile():
  def __init__(self,val):
    self.val=val
    self.x_offset=0
    self.y_offset=0
    self.max_xo=0
    self.max_yo=0
    self.new=True
    self.size=53

  def __add__(self,p):
    return self(self.x+p.x,self.y+p.y)

  def __str__(self):
    return '[{0}]'.format(2**self.val)

  def __eq__(t1,t2):
    try:
      return t1.val==t2.val
    except AttributeError:
      return False

  def clone(self):
    copy=Tile(self.val)
    copy.x_offset=self.x_offset
    copy.y_offset=self.y_offset
    copy.max_xo=self.max_xo
    copy.max_yo=self.max_yo
    copy.new=self.new
    return copy

  def draw(self,x,y):
    x1=59*x+5+self.x_offset+26-self.size//2
    y1=59*y+5+self.y_offset+26-self.size//2
    x2=x1+self.size-1
    y2=y1+self.size-1
    c=colors[self.val-1]
    fillrect(1,x1+1,y1+1,self.size-2,self.size-2,c,c)
    line(1,x1+3,y1,x2-3,y1,c)
    line(1,x1+3,y2,x2-3,y2,c)
    line(1,x1,y1+3,x1,y2-3,c)
    line(1,x2,y1+3,x2,y2-3,c)
    dimgrob(3,60,60,0)
    tx=x1+self.size//2-eval('TEXTOUT_P({0},G3,0,0,7)'.format(2**self.val))//2
    ty=y1+self.size//2-13
    eval('TEXTOUT_P({2},G1,{0},{1},7,#{3})'.format(tx,ty,2**self.val,'776E65' if self.val<3 else 'F9F6F2'))

class Board():
  def __init__(self):
    self.grid =[[0,0,0,0],[0,0,0,0],[0,0,0,0],[0,0,0,0]]
    self.temp1=[[0,0,0,0],[0,0,0,0],[0,0,0,0],[0,0,0,0]]
    self.temp2=[[0,0,0,0],[0,0,0,0],[0,0,0,0],[0,0,0,0]]
    self.place_rand()
    self.place_rand()
    for r in self.grid:
      for t in r:
        try:
          t.new=False
        except AttributeError:
          pass
    self.grids=(self.temp1,self.temp2,self.grid)
    self.score=0

  def __str__(self):
    return gridstr(self.grid)

  def __eq__(b1,b2):
    for j in range(4):
      for i in range(4):
        if b1.grid[j][i]!=b2.grid[j][i]:
          return False
    return True
    

  def draw(self,draw_new):
    for grid in self.grids:
      for i in range(4):
        for j in range(4):
          tile=grid[i][j]
          if tile!=0 and not (tile.new and not draw_new):
            tile.draw(j,i)
            #if tile.new and draw_new:
            #  print('e')

  def place_tile(self,x,y,v):
    self.grid[y][x]=Tile(v)

  def place_rand(self):
    x=randint(0,3)
    y=randint(0,3)
    if self.grid[y][x]!=0:
      try:
        self.place_rand()
      except RuntimeError:
        return
      return
    if randint(0,9):
      self.place_tile(x,y,1)
    else:
      self.place_tile(x,y,2)

  def clone(self):
    copy=Board()
    for j in range(4):
      for i in range(4):
        if self.grid[j][i]==0:
          copy.grid[j][i]=0
        else:
          copy.grid[j][i]=self.grid[j][i].clone()
    copy.score=self.score
    return copy

  def process_up(self):
    for i in range(4):
      k=0
      for j in range(4):
        if self.grid[j][i]!=0:
          self.grid[k][i]=self.grid[j][i]
          self.grid[k][i].y_offset=59*(j-k)
          k+=1
      for j in range(k,4):
        self.grid[j][i]=0
      for j in range(3):
        if self.grid[j][i]!=0 and self.grid[j+1][i]!=0 and self.grid[j][i].val==self.grid[j+1][i].val:
          self.temp1[j][i]=self.grid[j+1][i].clone()
          self.temp1[j][i].y_offset+=59
          self.temp2[j][i]=self.grid[j][i].clone()
          self.grid[j+1][i]=0
          self.grid[j][i].val+=1
          self.grid[j][i].new=True
          self.score+=2**self.grid[j][i].val
      k=0
      for j in range(4):
        if self.grid[j][i]!=0:
          self.grid[k][i]=self.grid[j][i]
          self.grid[k][i].y_offset+=59*(j-k)
          k+=1
      for j in range(k,4):
        self.grid[j][i]=0
    return self

  def process_down(self):
    for i in range(4):
      k=3
      for j in range(3,-1,-1):
        if self.grid[j][i]!=0:
          self.grid[k][i]=self.grid[j][i]
          self.grid[k][i].y_offset=59*(j-k)
          k-=1
      for j in range(k,-1,-1):
        self.grid[j][i]=0
      for j in range(3,0,-1):
        if self.grid[j][i]!=0 and self.grid[j-1][i]!=0 and self.grid[j][i].val==self.grid[j-1][i].val:
          self.temp1[j][i]=self.grid[j-1][i].clone()
          self.temp1[j][i].y_offset-=59
          self.temp2[j][i]=self.grid[j][i].clone()
          self.grid[j-1][i]=0
          self.grid[j][i].val+=1
          self.grid[j][i].new=True
          self.score+=2**self.grid[j][i].val
      k=3
      for j in range(3,-1,-1):
        if self.grid[j][i]!=0:
          self.grid[k][i]=self.grid[j][i]
          self.grid[k][i].y_offset+=59*(j-k)
          k-=1
      for j in range(k,-1,-1):
        self.grid[j][i]=0
    return self

  def process_left(self):
    for i in range(4):
      k=0
      for j in range(4):
        if self.grid[i][j]!=0:
          self.grid[i][k]=self.grid[i][j]
          self.grid[i][k].x_offset=59*(j-k)
          k+=1
      for j in range(k,4):
        self.grid[i][j]=0
      for j in range(3):
        if self.grid[i][j]!=0 and self.grid[i][j+1]!=0 and self.grid[i][j].val==self.grid[i][j+1].val:
          self.temp1[i][j]=self.grid[i][j+1].clone()
          self.temp1[i][j].x_offset+=59
          self.temp2[i][j]=self.grid[i][j].clone()
          self.grid[i][j+1]=0
          self.grid[i][j].val+=1
          self.grid[i][j].new=True
          self.score+=2**self.grid[i][j].val
      k=0
      for j in range(4):
        if self.grid[i][j]!=0:
          self.grid[i][k]=self.grid[i][j]
          self.grid[i][k].x_offset+=59*(j-k)
          k+=1
      for j in range(k,4):
        self.grid[i][j]=0
    return self

  def process_right(self):
    for i in range(4):
      k=3
      for j in range(3,-1,-1):
        if self.grid[i][j]!=0:
          self.grid[i][k]=self.grid[i][j]
          self.grid[i][k].x_offset=59*(j-k)
          k-=1
      for j in range(k,-1,-1):
        self.grid[i][j]=0
      for j in range(3,0,-1):
        if self.grid[i][j]!=0 and self.grid[i][j-1]!=0 and self.grid[i][j].val==self.grid[i][j-1].val:
          self.temp1[i][j]=self.grid[i][j-1].clone()
          self.temp1[i][j].x_offset-=59
          self.temp2[i][j]=self.grid[i][j].clone()
          self.grid[i][j-1]=0
          self.grid[i][j].val+=1
          self.grid[i][j].new=True
          self.score+=2**self.grid[i][j].val
      k=3
      for j in range(3,-1,-1):
        if self.grid[i][j]!=0:
          self.grid[i][k]=self.grid[i][j]
          self.grid[i][k].x_offset+=59*(j-k)
          k-=1
      for j in range(k,-1,-1):
        self.grid[i][j]=0
    return self

  def animate(self):
    for g in self.grids:
      for i in range(4):
        for j in range(4):
          t=g[i][j]
          if t!=0:
            t.max_xo=t.x_offset
            t.max_yo=t.y_offset
    lt=int(eval('ticks'))
    render()
    dt=int(eval('ticks'))-lt
    lt=int(eval('ticks'))
    st=lt
    while lt<st+125:
      for g in self.grids:
        for i in range(4):
          for j in range(4):
            t=g[i][j]
            if t!=0:
              sx=copysign(max(0,dt*abs(t.max_xo)/100),t.max_xo)
              sy=copysign(max(0,dt*abs(t.max_yo)/100),t.max_yo)
              if sx>0:
                t.x_offset=max(0,t.x_offset-sx)
              else:
                t.x_offset=min(0,t.x_offset-sx)
              if sy>0:
                t.y_offset=max(0,t.y_offset-sy)
              else:
                t.y_offset=min(0,t.y_offset-sy)
      render()
      dt=int(eval('ticks'))-lt
      lt=int(eval('ticks'))

    lt=int(eval('ticks'))
    st=lt
    while lt<st+150:
      f=lambda x: (4-(0.9*x-2)**2)*(100/117)*53/3
      for r in self.grid:
        for t in r:
          if t!=0 and t.new:
            x=min(3*(lt-st)/125,3)
            t.size=int(f(x))
      render(draw_new=True)
      dt=int(eval('ticks'))-lt
      lt=int(eval('ticks'))
    for i in range(4):
      for j in range(4):
        if self.grid[i][j]!=0:
          self.grid[i][j].new=False
        self.temp1[i][j]=0
        self.temp2[i][j]=0
    render()

def get_highscore():
  hs=0
  try:
    with open('highscore.dat','rb') as f:
      b=bytearray(2)
      f.readinto(b)
      hs=b[0]*256+b[1]
  except OSError:
    pass
  return hs

def set_highscore(s):
  with open('highscore.dat','wb') as f:
    b=bytearray((s//256,s%256))
    f.write(b)

def init_graphics():
  dimgrob(1,320,240,0)
  dimgrob(2,320,240,0xf9f6f2)
  fillrect(2,0,0,240,240,0xbbada0,0xbbada0)
  for i in range(4):
    for j in range(4):
      fillrect(2,59*j+4,59*i+4,55,55,0xccc0b4,0xccc0b4)
  eval('textout_p("Score:",G2,245,0,5,#776E65)')
  eval('textout_p("Highscore:",G2,245,40,4,#776E65)')
  eval('textout_p({0},G2,245,58,5,#776E65)'.format(get_highscore()))

def render(overlay=Overlay(0),draw_new=False):
  #print(overlay)
  blit(1,0,0,2)
  eval('textout_p({0},G1,245,20,5,#776E65)'.format(board.score))
  board.draw(draw_new)
  overlay.draw()
  blit(0,0,0,1)

def take_turn():
  while eval('getkey')!=-1 or len(eval('mouse')[0]):
    pass
  while True:
    render()
    key=eval('getkey')
    m=[(n+2**63)%(2**64)-2**63 for n in (eval('mouse')[0])]
    if len(m)>0 and m[4]==2:
      dx=m[0]-m[2]
      dy=m[1]-m[3]
      mdir='x' if abs(dx)>abs(dy) else 'y'
      #print(dx,dy,mdir)
    else:
      dx=0
      dy=0
      mdir=None
    if key==2 or (mdir=='y' and dy<0):
      if board.clone()!=board.clone().process_up():
        board.process_up()
        break
    if key==12 or (mdir=='y' and dy>0):
      if board.clone()!=board.clone().process_down():
        board.process_down()
        break
    if key==7 or (mdir=='x' and dx<0):
      if board.clone()!=board.clone().process_left():
        board.process_left()
        break
    if key==8 or (mdir=='x' and dx>0):
      if board.clone()!=board.clone().process_right():
        board.process_right()
        break

def game_over():
  if board.score>get_highscore():
    set_highscore(board.score)
  lt=int(eval('ticks'))
  render()
  dt=int(eval('ticks'))-lt
  lt=int(eval('ticks'))
  st=lt
  while lt<st+1000:
    render(GameOver(255*(lt-st)/1000))
    dt=int(eval('ticks'))-lt
    lt=int(eval('ticks'))

while True:
  init_graphics()
  board=Board()
  while True:
    render()
    take_turn()
    board.place_rand()
    board.animate()
    if board.clone()==board.clone().process_up() and board.clone()==board.clone().process_down() and board.clone()==board.clone().process_left() and board.clone()==board.clone().process_right():
      break
  game_over()
  while eval('getkey')!=-1 or len(eval('mouse')[0]):
    pass
  eval('wait(-1)')